/*
  f-cpu/c/scheduler/scheduler.c
  c simulation of the F-CPU decoder unit
  Copyright (C) 2002 Jaap Stolk (JWS) jwstolk@yahoo.com
  version:
           19 July 2002 13:30
Sun Jul 21 12:56:13 CEST 2002 JWS: added write queue.
Sun Jul 21 15:25:43 CEST 2002 JWS: romoved IMUL counter remarks (only IDIV)
Mon Jul 22 15:07:09 CEST 2002 JWS: complete rewrite, along with other units
Mon Jul 22 23:02:02 CEST 2002 JWS: finisched major scheduling update
Tue Jul 23 12:46:35 CEST 2002 JWS: normal bypass now overwrites delayed bypass
Thu Jul 25 21:52:20 CEST 2002 JWS: added 0-cycle moves remarks but no code :) 
Sun Jul 28 11:24:35 CEST 2002 JWS: added MAX_EU_LATENCY
Mon Jul 29 10:47:38 CEST 2002 JWS: added display_color and display_normal
Mon Jul 29 11:00:47 CEST 2002 JWS: used MAX_EU_LATENCY for clock queue
Mon Jul 29 13:46:40 CEST 2002 JWS: split dec_latency into _w0 and _w1
Tue Jul 30 14:30:01 CEST 2002 JWS: removed bug in bypass detection (_w0/1)
Fri Aug  2 17:25:00 CEST 2002 JWS: added JMP support
Fri Aug  2 21:04:16 CEST 2002 JWS: removed "register in use" tabel
Fri Aug  2 21:04:16 CEST 2002 JWS: added writeback dependency, and more

 ------------------------BEGIN-LICENSE------------------------------------
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 ---------------------------END-LICENSE-----------------------------------

     sch_operand_1  (from fetcher)
     sch_operand_2
     sch_operand_dst
         | | |
         v v v
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %%                         scheduler                                %%
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
         | | |
         v v v
     sch_bypass_r0 (no bypass, (direct/delayed) bypas to w0/w1
     sch_bypass_r1       
     sch_bypass_r2    (to the xbar)     

     sch_use_r0  (bool)  from the decoder
     sch_use_r1  (bool)
     sch_use_r2  (bool)
     sch_r0_reg_nr       from the decoder 
     sch_r1_reg_nr
     sch_r2_reg_nr
     sch_latency_w0
     sch_latency_w1
     sch_w0_port_nr
     sch_w1_port_nr
     sch_w0_reg_nr
     sch_w1_reg_nr
     | | | | | | | | | | | |
     v v v v v v v v v v v v
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %%                         scheduler                                %%
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
     | | | | | | |
     v v v v v v v
     stalled (bool)   to evrywhere !!!
     sch_xbar_w0_reg_nr   to x-bar
     sch_xbar_w0_port_nr
     sch_xbar_w1_reg_nr
     sch_xbar_w1_port_nr
     sch_reg_w0_reg_nr    to registers
     sch_reg_w1_reg_nr

  show the the exact reason for the stall, for debugging and optimising ?
*/

/* defines the constants, inputs and outputs : */
#include <scheduler.h>




void scheduler_view(void){
  int reg_view_cnt;
  int level_view_cnt;

  printf("scheduler inputs/outputs for bypass detection:\n");
  printf("sch_operand_1     =%s%2i%s  ",   display_color, sch_operand_1,   display_normal );
  printf("sch_operand_2     =%s%2i%s  ",   display_color, sch_operand_2,   display_normal );
  printf("sch_operand_dst    =%s%2i%s\n",  display_color, sch_operand_dst, display_normal );
  printf("(tmp_sch_operand_1=%s%2i%s  ",   display_color, sch_operand_1,   display_normal );
  printf("tmp_sch_operand_2 =%s%2i%s  ",   display_color, sch_operand_2,   display_normal );
  printf("tmp_sch_operand_dst=%s%2i%s)\n", display_color, sch_operand_dst, display_normal );
  printf("sch_bypass_r0     =%s%2X%s  ",   display_color, sch_bypass_r0,   display_normal );
  printf("sch_bypass_r1     =%s%2X%s  ",   display_color, sch_bypass_r1,   display_normal );
  printf("sch_bypass_r2      =%s%2X%s\n",  display_color, sch_bypass_r2,   display_normal );

  printf("   internal write queue:       0  1  2  3  4  5  6  7  8  9 10\n");

for (reg_view_cnt=0; reg_view_cnt<2; reg_view_cnt++){

  printf("sch_Q_write_slot_used   [0] = ");
  for (level_view_cnt=0;level_view_cnt<=(MAX_EU_LATENCY+2);level_view_cnt++){
    if ( level_view_cnt != 0 ) {
      if ( sch_Q_write_slot_used[level_view_cnt][reg_view_cnt] ) {
        printf(" %sX%s ", display_color, display_normal);
      }else{
        printf(" %s.%s ", display_color, display_normal);
      }
    }else{
      printf("   ");
    }
  }
  printf("\n");

  printf("sch_Q_write_to_reg_nr   [%1d] = ",reg_view_cnt);
  for (level_view_cnt=0;level_view_cnt<=(MAX_EU_LATENCY+2);level_view_cnt++){
    if ( level_view_cnt != 0 ) {
      printf("%s%2d%s ", display_color, sch_Q_write_to_reg_nr[level_view_cnt][reg_view_cnt], display_normal);
    }else{
      printf("   ");
    }
  }
  printf("\n");

  printf("sch_Q_write_from_port_nr[%1d] = ",reg_view_cnt);
  for (level_view_cnt=0;level_view_cnt<=(MAX_EU_LATENCY+2);level_view_cnt++){
    if ( level_view_cnt >2 ) {
      printf("%s%2d%s ", display_color, sch_Q_write_from_port_nr[level_view_cnt][reg_view_cnt], display_normal);
    }else{
      printf("   ");
    }
  }
  printf("\n");

}
  /* note: level 7 and 8 are the same when this is shown. (just after the shift) ??????*/

  printf("scheduler inputs:\n");
  printf("sch_use_r0     = %s  ",        display_bool[sch_use_r0 ] );
  printf("sch_use_r1     = %s  ",        display_bool[sch_use_r1 ] );
  printf("sch_use_r2     = %s\n",        display_bool[sch_use_r2 ] );
  printf("sch_r0_reg_nr  = %s%2i%s     ",display_color, sch_r0_reg_nr,   display_normal );
  printf("sch_r1_reg_nr  = %s%2i%s     ",display_color, sch_r1_reg_nr,   display_normal );
  printf("sch_r2_reg_nr  = %s%2i%s\n",   display_color, sch_r2_reg_nr,   display_normal );
  printf("sch_w0_port_nr = %s%2X%s     ",display_color, sch_w0_port_nr,  display_normal );
  printf("sch_w1_port_nr = %s%2X%s\n",   display_color, sch_w1_port_nr,  display_normal );
  printf("sch_w0_reg_nr  = %s%2X%s     ",display_color, sch_w0_reg_nr,   display_normal );
  printf("sch_w1_reg_nr  = %s%2X%s\n",   display_color, sch_w1_reg_nr,   display_normal );
  printf("scheduler outputs:        ");
  printf("stalled = %s\n",               display_bool[stalled ] );
  printf("sch_xbar_w0_reg_nr  = %s%2i%s  ", display_color, sch_xbar_w0_reg_nr, display_normal );
  printf("sch_xbar_w1_reg_nr  = %s%2i%s\n", display_color, sch_xbar_w1_reg_nr, display_normal );
  printf("sch_xbar_w0_port_nr = %s%2i%s  ", display_color, sch_xbar_w0_port_nr,display_normal );
  printf("sch_xbar_w1_port_nr = %s%2i%s\n", display_color, sch_xbar_w1_port_nr,display_normal );
  printf("sch_reg_w0_reg_nr   = %s%2i%s  ", display_color, sch_reg_w0_reg_nr,  display_normal );
  printf("sch_reg_w1_reg_nr   = %s%2i%s\n", display_color, sch_reg_w1_reg_nr,  display_normal );

}


static inline void scheduler_cycle (void) {
 
  int level_loop;
  int reg_loop;
  bool swapped;
  bool r0_busy;
  bool r1_busy;
  bool r2_busy;
  bool w0_busy;
  bool w1_busy;
  bool w0_on_w0_slot_used;
  bool w1_on_w1_slot_used;
  bool w0_on_w1_slot_used;
  bool w1_on_w0_slot_used;
  bool ports_used;

  /* FIXME: */
 bool  sch_use_w0;
 bool  sch_use_w1;


/*%%%%%%%%%%%%%%%% do a register write if it's indecated by the queue: %%%%%%*/
  /* (look at "level 2" in the queue) */

  sch_reg_w0_reg_nr = 0; /* --> 0 =  no write */
  sch_reg_w1_reg_nr = 0; /* --> 0 =  no write */

  if ( sch_Q_write_to_reg_nr[2][0] ){
    sch_reg_w0_reg_nr = sch_Q_write_to_reg_nr   [2][0];
  }
  if ( sch_Q_write_to_reg_nr[2][1] ){
    sch_reg_w1_reg_nr = sch_Q_write_to_reg_nr   [2][1];
  }

/*%%%%%%%%%%%%%%%% detect a stall: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/

  /* FIXME: */
  sch_use_w0 = sch_w0_reg_nr;
  sch_use_w1 = sch_w1_reg_nr;


  /* check if all the input registers are free: */
  r0_busy = false;
  r1_busy = false;
  r2_busy = false;
  for (level_loop=3; level_loop<=(MAX_EU_LATENCY+2); level_loop++){
    for (reg_loop=0; reg_loop<2; reg_loop++){
      if ( sch_Q_write_to_reg_nr[level_loop][reg_loop] == sch_r0_reg_nr ) r0_busy = true;
      if ( sch_Q_write_to_reg_nr[level_loop][reg_loop] == sch_r1_reg_nr ) r1_busy = true;
      if ( sch_Q_write_to_reg_nr[level_loop][reg_loop] == sch_r2_reg_nr ) r2_busy = true;
    }
  }
  if ( !sch_use_r0 ) r0_busy = false; /* disregard check */
  if ( !sch_use_r1 ) r1_busy = false;
  if ( !sch_use_r2 ) r2_busy = false;

  /* check for write dependency:  (do net check for writes BEFORE our own latency) */

  w0_busy = false;
  for (level_loop=3+sch_latency_w0; level_loop<=(MAX_EU_LATENCY+2); level_loop++){
    if ( sch_Q_write_to_reg_nr[level_loop][0] == sch_w0_reg_nr ) w0_busy = true;
  }
  if ( !sch_use_w0 ) w0_busy = false; /* disregard check */

  w1_busy = false;
  for (level_loop=3+sch_latency_w1; level_loop<=(MAX_EU_LATENCY+2); level_loop++){
    if ( sch_Q_write_to_reg_nr[level_loop][1] == sch_w1_reg_nr ) w1_busy = true;
  }
  if ( !sch_use_w1 ) w1_busy = false; /* disregard check */


  /* check if the needed write ports are free: */
  /* the nop instruction needs 0 free write ports !! */
  /* also check if swaped ports is posible. */
  w0_on_w0_slot_used = sch_Q_write_slot_used[3+sch_latency_w0][0]; /* normal */
  w1_on_w1_slot_used = sch_Q_write_slot_used[3+sch_latency_w1][1]; /* normal */
  w0_on_w1_slot_used = sch_Q_write_slot_used[3+sch_latency_w0][1]; /* swaped */
  w1_on_w0_slot_used = sch_Q_write_slot_used[3+sch_latency_w1][0]; /* swaped */
  swapped = false;

  if ( !sch_use_w0 ){
    w0_on_w0_slot_used = false; /* disregard check */
    w0_on_w1_slot_used = false;
  }
  if ( !sch_use_w1 ){
    w1_on_w0_slot_used = false; /* disregard check */
    w1_on_w1_slot_used = false;
  }

  if ( sch_use_w0 && !sch_use_w1 ){
    if (w0_on_w0_slot_used && !w0_on_w1_slot_used ) swapped = true;
  }
  if ( !sch_use_w0 && sch_use_w1 ){
    if (w1_on_w1_slot_used && !w1_on_w0_slot_used ) swapped = true;
  }

  if ( swapped ){
    ports_used = ( w0_on_w1_slot_used || w1_on_w0_slot_used );
  }else{
    ports_used = ( w0_on_w0_slot_used || w1_on_w1_slot_used );
  }

  stalled = r0_busy || r1_busy || r2_busy || w0_busy || w1_busy || ports_used;
  stalled = false;

/*%%%%%%%%%%%%%%%% add instruction to score board and write queue: %%%%%%%%%%*/
      /* if sch_latency==0, we don't need it in the port nr queue, it will we    */
      /*  removed from the queue anyway, when we clock the write port nr queue ?! */


  /* clear the toplevel of the queues: */
  for (reg_loop=0; reg_loop<2; reg_loop++){
    sch_Q_write_slot_used   [MAX_EU_LATENCY+2][reg_loop] = false;
    sch_Q_write_to_reg_nr   [MAX_EU_LATENCY+2][reg_loop] = 0; /* -> no clear needed !*/
    sch_Q_write_from_port_nr[MAX_EU_LATENCY+2][reg_loop] = 0; /* -> no clear needed !*/
  }

  /* add IDIV counter to toplevel when it reaches level 8 !? */

  if ( !stalled ){

    /* (no set counters for IDIV yet!!) */

    if ( swapped ){
      if (sch_use_w0){
        sch_Q_write_slot_used   [sch_latency_w0+3][1] = true;
        sch_Q_write_to_reg_nr   [sch_latency_w0+3][1] = sch_w0_reg_nr;
        sch_Q_write_from_port_nr[sch_latency_w0+3][1] = sch_w0_port_nr;
      }
      if (sch_use_w1){
        sch_Q_write_slot_used   [sch_latency_w1+3][0] = true;
        sch_Q_write_to_reg_nr   [sch_latency_w1+3][0] = sch_w1_reg_nr;
        sch_Q_write_from_port_nr[sch_latency_w1+3][0] = sch_w1_port_nr;
      }
    }else{
      if (sch_use_w0){
        sch_Q_write_slot_used   [sch_latency_w0+3][0] = true;
        sch_Q_write_to_reg_nr   [sch_latency_w0+3][0] = sch_w0_reg_nr;
        sch_Q_write_from_port_nr[sch_latency_w0+3][0] = sch_w0_port_nr;
      }
      if (sch_use_w1){
        sch_Q_write_slot_used   [sch_latency_w1+3][1] = true;
        sch_Q_write_to_reg_nr   [sch_latency_w1+3][1] = sch_w1_reg_nr;
        sch_Q_write_from_port_nr[sch_latency_w1+3][1] = sch_w1_port_nr;
      }
    }
  }

/*%%%%%%%%%%%%%%%% do an x-bar write if it's indecated by the queue: %%%%%%%%*/
  /* AFTER updating the queue, so we can do 0-cycle register move !! */
  /* (look at "level 3" in the queue) */

  /* instruct xbar to do a write from the EU output to a register: */
  sch_xbar_w0_reg_nr  = 0; /* --> 0 =  no write */
  sch_xbar_w0_port_nr = 0; /* --> not needed */
  sch_xbar_w1_reg_nr  = 0; /* --> 0 =  no write */
  sch_xbar_w1_port_nr = 0; /* --> not needed */
  if ( sch_Q_write_to_reg_nr[3][0] ){
    sch_xbar_w0_reg_nr  = sch_Q_write_to_reg_nr   [3][0];
    sch_xbar_w0_port_nr = sch_Q_write_from_port_nr[3][0];
  }
  if ( sch_Q_write_to_reg_nr[3][1] ){
    sch_xbar_w1_reg_nr  = sch_Q_write_to_reg_nr   [3][1];
    sch_xbar_w1_port_nr = sch_Q_write_from_port_nr[3][1];
  }

/*%%%%%%%%%%%%%%%% detect register bypass: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/

/*Get the 3 operands from the fetcher and find out if the instruction
  that is now being decoded, could need a register bypass.
  Ther are 5 possible results:
    - PORT_READ_FROM_REGISTER      (no bypass)
    - PORT_READ_FROM_BYPASS_0      (bypass from write bus 0)
    - PORT_READ_FROM_BYPASS_0_BIS  (1-cycle delay bypass from write bus 0)
    - PORT_READ_FROM_BYPASS_1      (bypass from write bus 1)
    - PORT_READ_FROM_BYPASS_1_BIS  (1-cycle delay bypass from write bus 1)
  Watch out: at this point it is stil unclear if the operands were actualy a
  register number ! doing a bypass if the operand was (part of) flags, could
  distroi 8/18bit imm data in the xbar bus, so check for that BEFORE doing
  the bypass. ???
  Watch out: no bypass dous NOT mean that the register value is ok. a
  pipeline stall could still be needed.
  After the stal cycle a bypass is usualy needed, but the stall could also
  have a different cause, like waiting for a free write slot.
  Watch out: if there is a write-after-write both a normal and a delayd bypass
  is posible !! we should use the normal bypass ! (not the delayed one ) */

/*If a xbar wrtite is on the bottom of the queue, it will be performed
  on the next cycle, witch is when we need a bypass.
  Please note: level 3 is the bottom of the queue (normal use) !!
  operand_1 = R0,  operand_2 = R1,  operand_dst = R2
  watch register 0 ! (hardwired to 0) */

  sch_bypass_r0 = PORT_READ_FROM_REGISTER; /* if no reg nr match is found */
  sch_bypass_r1 = PORT_READ_FROM_REGISTER;
  sch_bypass_r2 = PORT_READ_FROM_REGISTER;

  /* need to know stalled here !!!! */
  if ( stalled ){  /* bypass detection after a stall */
    sch_operand_1  = tmp_sch_operand_1;   /* ignor what we get from the  */
    sch_operand_2  = tmp_sch_operand_2;   /* fetcher and use the old     */
    sch_operand_dst= tmp_sch_operand_dst; /* values of the stalled instr.*/
  }

  /* check for bypass of read bus 0: */
  if ( sch_operand_1 ){
    if ( sch_operand_1  == sch_Q_write_to_reg_nr[1][0] ) sch_bypass_r0 = PORT_READ_FROM_BYPASS_0_BI2;
    if ( sch_operand_1  == sch_Q_write_to_reg_nr[2][0] ) sch_bypass_r0 = PORT_READ_FROM_BYPASS_0_BIS;
    if ( sch_operand_1  == sch_Q_write_to_reg_nr[3][0] ) sch_bypass_r0 = PORT_READ_FROM_BYPASS_0;
    if ( sch_operand_1  == sch_Q_write_to_reg_nr[1][1] ) sch_bypass_r0 = PORT_READ_FROM_BYPASS_1_BI2;
    if ( sch_operand_1  == sch_Q_write_to_reg_nr[2][1] ) sch_bypass_r0 = PORT_READ_FROM_BYPASS_1_BIS;
    if ( sch_operand_1  == sch_Q_write_to_reg_nr[3][1] ) sch_bypass_r0 = PORT_READ_FROM_BYPASS_1;
  }

  /* check for bypass of read bus 1: */
  if ( sch_operand_2 ){
    if ( sch_operand_2  == sch_Q_write_to_reg_nr[1][0] ) sch_bypass_r1 = PORT_READ_FROM_BYPASS_0_BI2;
    if ( sch_operand_2  == sch_Q_write_to_reg_nr[2][0] ) sch_bypass_r1 = PORT_READ_FROM_BYPASS_0_BIS;
    if ( sch_operand_2  == sch_Q_write_to_reg_nr[3][0] ) sch_bypass_r1 = PORT_READ_FROM_BYPASS_0;
    if ( sch_operand_2  == sch_Q_write_to_reg_nr[1][1] ) sch_bypass_r1 = PORT_READ_FROM_BYPASS_1_BI2;
    if ( sch_operand_2  == sch_Q_write_to_reg_nr[2][1] ) sch_bypass_r1 = PORT_READ_FROM_BYPASS_1_BIS;
    if ( sch_operand_2  == sch_Q_write_to_reg_nr[3][1] ) sch_bypass_r1 = PORT_READ_FROM_BYPASS_1;
  }

  /* check for bypass of read bus 2: */
  if ( sch_operand_dst ){
    if ( sch_operand_dst== sch_Q_write_to_reg_nr[1][0] ) sch_bypass_r2 = PORT_READ_FROM_BYPASS_0_BI2;
    if ( sch_operand_dst== sch_Q_write_to_reg_nr[2][0] ) sch_bypass_r2 = PORT_READ_FROM_BYPASS_0_BIS;
    if ( sch_operand_dst== sch_Q_write_to_reg_nr[3][0] ) sch_bypass_r2 = PORT_READ_FROM_BYPASS_0;
    if ( sch_operand_dst== sch_Q_write_to_reg_nr[1][1] ) sch_bypass_r2 = PORT_READ_FROM_BYPASS_1_BI2;
    if ( sch_operand_dst== sch_Q_write_to_reg_nr[2][1] ) sch_bypass_r2 = PORT_READ_FROM_BYPASS_1_BIS;
    if ( sch_operand_dst== sch_Q_write_to_reg_nr[3][1] ) sch_bypass_r2 = PORT_READ_FROM_BYPASS_1;
  }


/*%%%%%%%%%%%%%%%% clock scoreboard (and clock write queue): %%%%%%%%%%%%%%%%*/

/*(in hardware this is done by the stage clock after each cycle)
  rotatong ptr's to levels could be used if this simulates to slow !!
  some FIFO's do not use level 0,1 or 2 !! */

  /* clock the sch_Q_write_slot_used FIFO */
  for (level_loop=1; level_loop<(MAX_EU_LATENCY+2); level_loop++){
    for (reg_loop=0; reg_loop<2; reg_loop++){
      sch_Q_write_slot_used   [level_loop  ][reg_loop] = 
      sch_Q_write_slot_used   [level_loop+1][reg_loop];
    }
  }

  /* clock the sch_Q_write_to_reg_nr FIFO */
  for (level_loop=1; level_loop<(MAX_EU_LATENCY+2); level_loop++){
    for (reg_loop=0; reg_loop<2; reg_loop++){
      sch_Q_write_to_reg_nr   [level_loop  ][reg_loop] = 
      sch_Q_write_to_reg_nr   [level_loop+1][reg_loop];
    }
  }

  /* clock the sch_Q_write_from_port_nr FIFO */
  for (level_loop=3; level_loop<(MAX_EU_LATENCY+2); level_loop++){
    for (reg_loop=0; reg_loop<2; reg_loop++){
      sch_Q_write_from_port_nr[level_loop  ][reg_loop] = 
      sch_Q_write_from_port_nr[level_loop+1][reg_loop];
    }
  }

/*%%%%%%%%%%%%%%%% delay operand data from fetcher: %%%%%%%%%%%%%%%%*/
  /* these are needed to detect a bypass if there was a stall !!! */

  if ( !stalled ){
    tmp_sch_operand_1  = sch_operand_1;
    tmp_sch_operand_2  = sch_operand_2;
    tmp_sch_operand_dst= sch_operand_dst;
  }

}



